#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define SHORT       0x1a
#define LONG        0x28
#define LONGLEAD    0x1000
#define SHORTLEAD   0x80
#define BLOCKWORDS  0x2d

static unsigned long length = 0;
static FILE *outfile;
static unsigned char cf = 0;

static inline unsigned char adc(unsigned char x, unsigned char y) {
    int z;
    z = (int)x + (int)y + (int)cf;
    cf = (z >> 8) & 1;
    return z & 0xff;
}

static inline unsigned char sbc(unsigned char x, unsigned char y) {
    int z;
    z = (int)x - (int)y - ((int)cf ^ 1);
    cf = ((z >> 8) & 1) ^ 1;
    return z & 0xff;
}

static inline unsigned char rol(unsigned char x) {
    unsigned char t;
    t = cf;
    cf = (x >> 7) & 1;
    return (x << 1) | t;
}

static inline unsigned char ror(unsigned char x) {
    unsigned char t;
    t = cf;
    cf = x & 1;
    return (x >> 1) | (t << 7);
}

static unsigned char rotate(unsigned char x, int n) {
    int i;
    if(n < 0)
        for(i=0; i<-n; i++) x = ror(x);
    else
        for(i=0; i<n; i++) x = rol(x);
    return x;
}

static void mix(unsigned char *data) {
    unsigned char a;
    int i;
    int pos[8] = { 2, 1, 3, 0, 3, 1, 0, 2 };
    int rot[8] = { -3, 3, -4, 2, 2, -3, -4, 0 };

    cf = 0;
    a = data[0] ^ 0xe2;

    for(i=0; i<8; i++) {
        a ^= data[pos[i]];
        a = rotate(a, rot[i]);
        a = adc(a, data[pos[i]]);
        data[pos[i]] = a;
    }
}

static void putbit(int c) {
    fputc(c?LONG:SHORT, outfile);
    length++;
}

static void putbyte(int c) {
    int i;
    for(i=0; i<8; i++) putbit((c<<i)&0x80);
}

#define SYNCLEN 5       /* Length in bytes of synchronization sequence */
#define CONFLEN 8       /* Length in bytes of confirmation sequence */

static void putsync(int n, int end) {
    int i;
    const unsigned char sync[SYNCLEN] =
        { 0x00, 0xa6, 0x8d, 0x95, 0x8e };
    const unsigned char conf[CONFLEN] =
        { 0xde, 0xd5, 0xf5, 0xe0, 0xc8, 0x4a, 0x93, 0x24 };

    for(i=0; i<n; i++) putbit(1);
    for(i=0; i<SYNCLEN; i++) putbyte(sync[i]);
    if(end)
        for(i=0; i<CONFLEN; i++) putbyte(conf[i]);
    else
        for(i=0; i<CONFLEN; i++) putbyte(conf[i] ^ 0xff);
}

static unsigned char hamming(const unsigned char *x) {
    int i;
    unsigned char a,b,c,parity;
    const unsigned char table[3][8] = {
        { 0x03, 0x05, 0x06, 0x07, 0x09, 0x0a, 0x0b, 0x0c },
        { 0x0d, 0x0e, 0x0f, 0x11, 0x12, 0x13, 0x14, 0x15 },
        { 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d }
    };

    a = ~x[0]; b = ~x[1]; c = ~x[2];
    parity = 0;
    for(i=0; i<8; i++) {
        if(!(a & 1)) parity ^= table[0][i];
        if(!(b & 1)) parity ^= table[1][i];
        if(!(c & 1)) parity ^= table[2][i];
        a = a >> 1; b = b >> 1; c = c >> 1;
    }

    return parity;
}

static void createblock(unsigned char *block, unsigned char *data, int adr) {
    unsigned char checksum[4];
    int i;

    checksum[0] = 0x00;
    checksum[1] = 0x00;
    checksum[2] = 0x00;
    checksum[3] = 0x00;

    for(i=0; i<0x80; i+=4) {
        checksum[0] ^= data[i+0];
        checksum[1] ^= data[i+1];
        checksum[2] ^= data[i+2];
        checksum[3] ^= data[i+3];
        mix(checksum);
    }

    memcpy(block, data, 0x80);

    block[0x80] = adr & 0xff;
    block[0x81] = (adr >> 8) & 0xff;
    block[0x82] = (adr >> 16) & 0xff;
/*
    block[0x80] = adr & 0xff;
    block[0x81] = (adr >> 8) & 0x3f;
    block[0x82] = adr >> 14;
*/

    checksum[0] ^= block[0x80];
    checksum[1] ^= block[0x81];
    checksum[2] ^= block[0x82];
    mix(checksum);

    memcpy(block+0x83, checksum, 4);
}

static void putblock(const unsigned char *x) {
    unsigned char parity[BLOCKWORDS];
    int i,j,k;

    putsync(SHORTLEAD, 0);

    for(i=0; i<BLOCKWORDS; i++) {
        parity[i] = hamming(x + (i*3));
    }

    k = 0;
    for(j=0; j<8; j++) {
        for(i=0; i<3*BLOCKWORDS; i++) {
            fputc(((x[k]<<j)&0x80)?LONG:SHORT, outfile);
            k+=3;
            if(k >= 3*BLOCKWORDS) k -= 3*BLOCKWORDS + 1;
            if(k < 0) k = 2;
            length++;
        }
    }
    for(j=0; j<5; j++) {
        for(i=0; i<BLOCKWORDS; i++) {
            fputc(((parity[i]<<j)&0x10)?LONG:SHORT, outfile);
            length++;
        }
    }
}

int main(int argc, char *argv[]) {
    unsigned char header[] =
        "C64-TAPE-RAW\x01\0\0\0\0\0\0\0\0\x40\x41\x0f";
    unsigned long addr = 0;
    unsigned char data[0x80], block[BLOCKWORDS*3];
    FILE *file;
    int i;

    if(argc != 3) {
        fprintf(stderr, "Usage: %s <infile> <outfile>\n", argv[0]);
        return 1;
    }
    if((file = fopen(argv[1], "r")) == NULL) {
        fprintf(stderr, "Can not open file '%s'\n", argv[1]);
        return 1;
    }
    if((outfile = fopen(argv[2], "w")) == NULL) {
        fprintf(stderr, "Can not open file '%s'\n", argv[2]);
        return 1;
    }

    length = 4;

    /* Write temporary header, before length is computed */
    fwrite(header, 24, 1, outfile);

    while(!feof(file)) {
        memset(data, 0, 0x80);
        fread(data, 1, 0x80, file);
        createblock(block, data, addr);
        addr++;
        putblock(block);
    }

    putsync(SHORTLEAD, 1);
    putbyte(0xff);

    rewind(outfile);
    for(i=0; i<4; i++) header[16+i] = length >> (8*i);
    fwrite(header, 24, 1, outfile);

    fclose(file);
    fclose(outfile);
    return 0;
}

